#---------------------------------------------------------------------------
#
#   MWG - Level View
#
#---------------------------------------------------------------------------

from __future__ import division
from math import pi
from pygame import Rect, draw
from albow.widget import Widget
from albow.image_array import get_image_array
from tiles import tile_images
from constants import *
from moving_objects import YarnBall
from vector_algebra import vadd

deg = pi / 180

yarn_map = {
	YARN_EW: 0,
	YARN_NS: 1,
	YARN_SE: 2,
	YARN_SW: 3,
	YARN_NW: 4,
	YARN_NE: 5,
}

yarn_tiles = get_image_array("yarn_segments.png", len(yarn_map), border = 2)
yarn_end_images = get_image_array("yarn_ends.png", 4, border = 2)

class LevelView(Widget):

	yarn_color = (255, 255, 0)

	window_size = (10, 9)  # in cells
	cell_size = (80, 60)    # in pixels
	
	show_cell_grid = False

	def __init__(self, game):
		ww, wh = self.window_size
		cw, ch = self.cell_size
		rect = Rect(0, 0, ww * cw, wh * ch)
		Widget.__init__(self, rect)
		self.game = game
	
	def level(self):
		return self.game.playing_level

	def get_window_pos(self):
		return self.level().window_pos
	
	def set_window_pos(self, p):
		self.level().window_pos = p
	
	def draw(self, surf):
		self.adjust_window()
		level = self.level()
		lw, lh = level.level_size
		ww, wh = self.window_size
		cw, ch = self.cell_size
		wx, wy = level.window_pos
		i0 = max(0, wx - 1)
		j0 = max(0, wy - 1)
		i1, j1 = min(wx + ww + 1, lw), min(wy + wh + 1, lh)
		cell_objects = [[[] for j in xrange(j0, j1)]
			for i in xrange(i0, i1)]
		for obj in level.moving_objects:
			i, j = obj.cell_coords()
			if i0 <= i < i1 and j0 <= j < j1:
				cell_objects[i - i0][j - j0].append(obj)
		def itercells():
			for i in xrange(i0, i1):
				r = Rect((i - wx) * cw, (j - wy) * ch, cw, ch)
				cell = level.cells[i][j]
				yield i, r, cell
		for j in xrange(j0, j1):
			for i, r, cell in itercells():
				self.draw_ground(surf, r, cell)
				self.draw_pavement(surf, r, cell)
				for obj in cell_objects[i - i0][j - j0]:
					if isinstance(obj, YarnBall):
						centre = self.object_coords(obj, wx, wy)
						self.draw_yarn_end(surf, r, obj, centre)
						if obj is cell.object_falling_in:
							obj.draw(surf, centre)
				if cell.object_falling_in:
					self.draw_ground_front(surf, r, cell)
				self.draw_yarn(surf, r, cell)
		for j in xrange(j0, j1):
			for i, r, cell in itercells():
				self.draw_item(surf, r, cell)
				self.post_draw_cell(surf, cell, r)
			objects = level.moving_objects_in_row(j)
			objects.sort(key = lambda obj: obj.position[1])
			for obj in objects:
				if not obj.falling_into_cell:
					obj.draw(surf, self.object_coords(obj, wx, wy))
					self.post_draw_object(surf, obj)
			for i, r, cell in itercells():
				self.draw_obstacle(surf, r, cell)
	
	def draw_ground(self, surf, r, cell):
		self.draw_tile(surf, r, cell.ground)
		if self.show_cell_grid:
			draw.rect(surf, self.grid_color, r, 1)
	
	def draw_pavement(self, surf, r, cell):
		self.draw_tile(surf, r, cell.pavement)
	
	def draw_yarn(self, surf, r, cell):
		for yarn in cell.yarn:
			index = yarn_map[yarn[:2]]
			image = yarn_tiles[index]
			surf.blit(image, r)
	
	def draw_item(self, surf, r, cell):
		self.draw_tile(surf, r, cell.item, 'center')
	
	def draw_obstacle(self, surf, r, cell):
		self.draw_tile(surf, r, cell.obstacle, 'bottomright')
	
	def draw_ground_front(self, surf, r, cell):
		ground = cell.ground
		if ground:
			self.draw_tile(surf, r, ground + "-front")
	
	def post_draw_object(self, surf, obj):
		pass
	
	def post_draw_cell(self, surf, cell, r):
		pass
	
	def object_coords(self, obj, i0, j0):
		cw, ch = self.cell_size
		px, py = obj.position
		x = int((px - i0) * cw)
		y = int((py - obj.z - j0) * ch)
		return (x, y)

	def draw_tile(self, surf, frame, name, align = 'midbottom'):
		if name:
			image = tile_images.get(name)
			if image:
				r = image.get_rect()
				setattr(r, align, getattr(frame, align))
				surf.blit(image, r)

	def draw_yarn_end(self, surf, cell_rect, obj, center):
		dir = obj.yarn_dir
		if dir is not None:
			#print "LevelView.draw_yarn_end:", cell_rect, center, dir ###
			cr = cell_rect
			cx, cy = center
			if dir == DIR_EAST:
				end = cr.midright
				w = 2 * (cr.right - cx)
				h = 2 * (cr.centery - cy)
				r = Rect(0, 0, w, abs(h))
				r.midleft = center
				if h > 0:
					a = (1, 2)
				else:
					a = (2, 3)
			elif dir == DIR_SOUTH:
				end = cr.midbottom
				w = 2 * (cr.centerx - cx)
				h = 2 * (cr.bottom - cy)
				r = Rect(0, 0, abs(w), h)
				r.midtop = center
				if w > 0:
					a = (3, 4)
				else:
					a = (2, 3)
			elif dir == DIR_WEST:
				end = cr.midleft
				w = 2 * (cx - cr.left)
				h = 2 * (cr.centery - cy)
				r = Rect(0, 0, w, abs(h))
				r.midright = center
				if h > 0:
					a = (0, 1)
				else:
					a = (3, 4)
			else: # DIR_NORTH
				end = cr.midtop
				w = 2 * (cr.centerx - cx)
				h = 2 * (cy - cr.top)
				r = Rect(0, 0, abs(w), h)
				r.midbottom = center
				if w > 0:
					a = (0, 1)
				else:
					a = (1, 2)
			#draw.rect(surf, (0, 0, 255), r) ###
			color = self.yarn_color
			thick = 4
			if r.height < 2 * thick:
				draw.line(surf, color, center, vadd(end, (0, -1)), thick)
			elif r.width < 2 * thick:
				draw.line(surf, color, center, vadd(end, (-1, 0)), thick)
			else:
				r.inflate_ip(thick, thick)
				ea = -90 * a[0] * deg
				sa = -90 * a[1] * deg
				#print "angles =", sa / (90 * deg), ea / (90 * deg) ###
				draw.arc(surf, color, r, sa, ea, thick)

	def adjust_window(self):
		wx, wy = self.get_window_pos()
		ww, wh = self.window_size
		x, y = self.level().player.position
		#d = 0.25
		d = 0.5
		x0 = int(x - d)
		y0 = int(y - d)
		x1 = int(x + d)
		y1 = int(y + d)
		#print "LevelView.adjust_window: player pos =", (x, y) ###
		#print "...buffered pos covers cells", (x0, x1), (y0, y1) ###
		#print "...current window =", (wx, wx + ww), (wy, wy + wh) ###
		if x0 < wx:
			wx = x - ww + 1
		elif x1 >= wx + ww:
			wx = x
		if y0 < wy:
			#print "...y0 < wy" ###
			wy = y - wh + 1
		elif y1 >= wy + wh:
			#print "...y1 >= wy + wh" ###
			wy = y
		self.move_window_to((int(wx), int(wy)))
	
	def move_window_to(self, (wx, wy)):
		lw, lh = self.level().level_size
		ww, wh = self.window_size
		wx = max(0, min(wx, lw - ww))
		wy = max(0, min(wy, lh - wh))
		#print "Level: moving window to", (wx, wy) ###
		self.set_window_pos((wx, wy))
